/* FILE: delta.cc          -*-Mode: c++-*-
 *
 * Micromagnetic hysteresis modeling
 *
 * History:
 *
 *      alpha: Began in spring 1995.
 *      beta:  August-October 1995; Complete code, including adaptive
 *                   stepsize and runtime selected magnetization
 *                   initialization, but only brute-force demag calculations,
 *                   and only for the "Pipe" model.  Also no precession terms.
 *      gamma: 25-October 1995; Incorporates RDM's FFT code to do demag
 *                   calculations.  Does both pipes and slab models.
 *                   Added seed initialization to UnifRand().  Added
 *                   uniaxial anisotropy (in addition to cubic).
 *      hyst:   1-November 1995;  Hysteresis modeling
 *              5-December 1995;  Added support for GenCubic MagElt type
 *              9-May      1996;  Added demag type selection
 *
 *      delta: 10-May 1996; Conversion of hyst to use mif-file input.
 *             30-Sep 1996; Support for non-uniform external fields
 *             17-Jul 1997; Improved signal handling and wrote self-
 *                          contained `tee` code.  -USR1 & -USR2 signals
 *                          (via `kill`) may be used to move the process
 *                          in and out of background (batch) mode.
 */

#include <fcntl.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <sys/times.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <time.h>
#include <unistd.h>

#include "misc.h"
#include "vecio.h"
#include "grid.h"
#include "mifio.h"
#include "zeeman.h"

#include "magelt.h"
#include "grid.h"

/* End includes */

#define MYBUFSIZE 1024
#define MYSMALLBUFSIZE 32

#undef TIMETEST
#ifdef TIMETEST
StopWatch tempwatch[5];
extern "C" { void mywatch(int timer,int start); }
void mywatch(int timer,int start)
{
  if(start) tempwatch[timer].Start();
  else      tempwatch[timer].Stop();
}
#endif // TIMETEST

void Usage(void)
{
  fprintf(stderr,"\n");
  fprintf(stderr,"Micromagnetic hysteresis modeling---\n");
  fprintf(stderr,"  USAGE: delta <filename.mif>\n\n");
  exit(1);
}

// stopflag condition codes
#define SF_NOSTOP      0
#define SF_TORQUE      1
#define SF_USERTORQUE  2
#define SF_USERSTOP    3
#define SF_PASSMAX     4

// Clean up zombies
#ifdef __GNUC__
void KillZombie(...)
#else
void KillZombie(int)
#endif
{
  while(waitpid(-1,NULL,WNOHANG)>0); // Kill any zombie'd children
  // Reset signal handlers
  signal(SIGCHLD,KillZombie);
#ifdef SIGCLD
  signal(SIGCLD,KillZombie);
#endif
}


// Background/interactive mode control
static int autodisplayflag=0; // 1=> Auto results display
static int backgroundflag=0;  // 1=> Running in background

void SetInteractiveSignalState();
void SetBackgroundSignalState();

#ifdef __GNUC__
void SetBackgroundMode(...)
#else
void SetBackgroundMode(int)
#endif
{ // Put into background processing mode
  // Set flags
  backgroundflag=1;
  autodisplayflag=0;
  // Reset signal handlers
  SetBackgroundSignalState();
}

#ifdef __GNUC__
void SetInteractiveMode(...)
#else
void SetInteractiveMode(int)
#endif
{ // Return to interactive processing mode
  // Set flags
  backgroundflag=0;
  autodisplayflag=0;  // Keep autodisplay turned off until
                      // user interactive re-activates it
  // Reset signal handlers
  SetInteractiveSignalState();
}

void SetInteractiveSignalState()
{
  // "Interactive" mode signal handler states
  SetupTerm();  RawTerm();

  // Setup signal handling for jumping to "background" mode
  signal(SIGHUP,SetBackgroundMode);
  signal(SIGTERM,SIG_IGN);
  signal(SIGTTIN,SetBackgroundMode);
  signal(SIGTTOU,SetBackgroundMode);
  signal(SIGUSR1,SetBackgroundMode);
  signal(SIGUSR2,SIG_IGN);
}

void SetBackgroundSignalState()
{
  signal(SIGINT,SIG_DFL);  signal(SIGQUIT,SIG_DFL);
  signal(SIGTSTP,SIG_DFL); signal(SIGCONT,SIG_DFL);

  signal(SIGHUP,SIG_IGN); // Ignore hangups (allows terminal logouts)
  signal(SIGTERM,SIG_IGN);
  signal(SIGTTIN,SIG_IGN); signal(SIGTTOU,SIG_IGN);
  signal(SIGUSR1,SIG_IGN);

  signal(SIGUSR2,SetInteractiveMode);  // Allow escape via SIGUSR2
}


void DisplayMagState(char* winname,Grid2D* grid,
		     int dolog2,FILE* log2file)
{
  pid_t display_pid;
  fflush(log2file); fflush(stdout); fflush(stderr);  // Safety
  if((display_pid=fork())==(pid_t)(-1))
    perror("Unable to spawn child display process (DisplayMagState())");
  else if(display_pid==(pid_t)0) { // Child process
    // Reset signal handlers
    signal(SIGINT,SIG_DFL);  signal(SIGQUIT,SIG_DFL);
    signal(SIGTSTP,SIG_IGN); signal(SIGCONT,SIG_DFL);
    if(dolog2) fclose(log2file);
    fclose(stdout); fclose(stdin);

    int Nx,Nz;
    grid->GetDimens(Nx,Nz);
    double **temparr=AllocateDoubles(Nx,3*Nz);
    grid->GetMag(temparr);
    VecWindow vw(Nz,Nx,temparr,3,0,2,1,winname);
    vw.CopyArray();
    DeallocateDoubles(temparr,Nx);
    delete grid;
    vw.Show();
    _exit(0);
  }
  else { // Main (parent) process
#if HAS_USLEEP
    usleep(500000); // Pause briefly to allow display in child
#else
    sleep(1);
#endif
  }
}

int main(int argc,char **argp)
{
  const double stop_torque=1e-4;      // Termination condition (torque)
  const double JiggleSize=PERTURBATION_SIZE;  // Bound on jiggle size
  const double JiggleTime=2;  // Number of times to jiggle before giving up

  TrackTime timer;
  int pass,stopflag;
  int keycode;
  char winname[MYBUFSIZE],filename[MYBUFSIZE];
  double last_torque;
  double timeindex;  // Elapsed time, in seconds
  double jiggle_count,jiggle_energy;
  double avemag_arr[VECDIM];
  ThreeVector avemag(avemag_arr);
  int passdumpflag=0;    // 1=> Dump intermediate state each report_interval
  MIF mif;
  int dolog2;
  AutoString LogName,Log2Name;
  FILE *logfile,*log2file;
  int bstep,bsteptotal;
  double bapparr[VECDIM]; ThreeVector bapp(bapparr);
  int Nx,Nz,report_interval,max_pass_count;
  double exch_energy,anis_energy,demag_energy,zeeman_energy,total_energy;
  char smallbuf[MYSMALLBUFSIZE];
  double dtemp;
  
  if(argc!=2) Usage();

  // Setup signal handler for child zombies
  signal(SIGCHLD,KillZombie);
#ifdef SIGCLD
  signal(SIGCLD,KillZombie);
#endif

  // "Interactive" mode signal handler states
  SetInteractiveSignalState();

  // SetupSignalLogger(); // For signal debugging /**/
  // signal(SIGTSTP,SIG_DFL); /**/

  time_t mytime;
  time(&mytime);

  mif.Read(argp[1]);

  UnifRand(mif.GetRandomizerSeed()); // Initialize random number generator
  report_interval=mif.GetUserReportCode();
  if(report_interval==0) report_interval=1;
  if(report_interval<0) {
    report_interval*=-1;
    passdumpflag=1;
  }
  max_pass_count=mif.GetMaxIterationCount();  // Note: <0 => No limit
  backgroundflag=0; autodisplayflag=1;
  if(mif.GetUserInteractionLevel()<1) SetBackgroundMode(1);
  else if(mif.GetUserInteractionLevel()<2)
    autodisplayflag=0;

  // Open logfile
  LogName.Dup(mif.GetOutBaseName());
  LogName.Append(".log");
  if((logfile=fopen(LogName.GetStr(),"a"))==NULL)
    fatalerr("main","Unable to open log file %s\n",LogName.GetStr());

  // Check to see if we want second log file.  If so, then set up
  // a `tee` process to duplicate everything written to log2file
  // to stdout.  If not, just set log2file to stdout.  (Basically,
  // log2file receives everything that goes to stdout, *except*
  // interactive user messages.)  With a copy-on-write fork, e.g.,
  // on a Linux system, this child process appears to only eat about
  // 70K of memory (on an ix86).  Of course, YMMV.
  dolog2=mif.GetLogLevel();
  log2file=stdout;  // Default; also used if the fork fails
  if(dolog2>0) {
    int fildes[2];
    pid_t tee_pid;
    if(pipe(fildes)!=0)
      fatalerr("main","Unable to open pipe for log file #2\n");
    fflush(stdout); fflush(stderr); // Safety
    if((tee_pid=fork())==(pid_t)(-1)) {
      perror("Unable to spawn child 'tee' (delta.cc: main())");
    }
    else if(tee_pid==(pid_t)0) {
      // Fork successful: Child process
      signal(SIGHUP,SIG_IGN);  // Protect against shell death
      signal(SIGTERM,SIG_IGN);

      signal(SIGTSTP,SIG_IGN); // Don't allow ^Z in parent
      signal(SIGSTOP,SIG_IGN); // to stop the child
      signal(SIGCONT,SIG_IGN);

      // Set up fildes usage
      int infileno=fildes[0];
      close(fildes[1]); // Not needed in child

      // Open output file
      Log2Name.Dup(mif.GetOutBaseName());
      Log2Name.Append(".log2");
      int dmpfileno=open(Log2Name.GetStr(),
			 O_WRONLY|O_CREAT|O_APPEND,
			 S_IREAD|S_IWRITE|S_IRGRP|S_IROTH);
      if(dmpfileno<0) {
	perror("main-Child");
	fatalerr("main-Child","Unable to open log file #2, %s\n",
		 Log2Name.GetStr());
      }
      int outfileno=fileno(stdout); fflush(stdout);

      // Set up i/o buffering
      int readcount;
      char mybuf[BUFSIZ];

      // Now just continuously copy from infile to dmpfile + stdout
      // NOTE: These don't need to be buffered, because the infileno
      //       pipe is buffered on the other (write) side.
      while((readcount=read(infileno,mybuf,BUFSIZ))>0) {
	// The above read should block until input is available
	write(dmpfileno,mybuf,readcount);
	write(outfileno,mybuf,readcount);
      }
      // This code gets executed when the parent closes fildes[1]
      close(dmpfileno);
      _exit(0);
    }
    else {
      // Fork successful: Parent process
      if((log2file=fdopen(fildes[1],"a"))==NULL)
	fatalerr("main-Parent",
		 "Unable to fdopen log file #2 pipe for output\n");
      setvbuf(log2file,NULL,_IOFBF,0);
      close(fildes[0]); // Not needed in parent
      signal(SIGPIPE,SIG_IGN);
      // Ignore broken pipes, just to make the parent more robust
      // in case the child dies.
    }
  }

  // Write first part of log file header
  fprintf(logfile,"# Start time: %s",ctime(&mytime));
  mif.DumpPartialInfo(logfile);
  fflush(logfile);
  fprintf(log2file,"# Start time: %s",ctime(&mytime));
  mif.DumpPartialInfo(log2file);
  fflush(log2file); fflush(stdout);

  // Initialize grid
  Grid2D *grid= new Grid2D(mif);
  grid->GetDimens(Nx,Nz);
  Zeeman *extfield=mif.GetExternalFieldType();
  grid->SethApplied(extfield);

  // Write remaining portion of logfile header
  fprintf(logfile,"#  Grid dimensions (x,z): %3d x %3d\n",Nx,Nz);
  fprintf(logfile,
          "#                                                 "
	  "                                      Max   Field \n"
          "#Point  Pass Time (ns)    Bx       By       Bz    "
	  "   Mx       My       Mz     Energy   Angle Updates\n");
  fflush(logfile);
  fprintf(log2file,"#  Grid dimensions (x,z): %3d x %3d\n",Nx,Nz);
  fflush(log2file); fflush(stdout);

  // Begin loops
  bsteptotal=mif.GetFieldTotalCount(); bstep=0;
  pass=0; timeindex=0.0;

  mif.ResetField();
  bapp.Dup(mif.GetNomExternalField());
  extfield->SetNomField(mif.GetMs(),bapp);
  grid->ResetODE(); // Initialize ODE solver
  jiggle_count=0; jiggle_energy=0.0;
  stopflag=SF_NOSTOP;
  last_torque=0.0; // Dummy value; NOTE: Value of last_torque is the
  // torque value calculated _before_ the most recent spin update.


  // Print initial field value
  fprintf(log2file,"# Initial applied field (T): (%8.5f,%8.5f,%8.5f)\n",
	 bapp[0],bapp[1],bapp[2]);

  // Print output columns super-header
  fprintf(log2file,"#                        Max   "
	 " -------------------- ENERGY --------------------"
	 "   Max   Field\n");

  last_torque=grid->GetMaxTorque();  // Initial configuration torque
  do {
    if(pass%(report_interval*20)==0) {
      if(pass!=0) {
	bapp.Dup(mif.GetNomExternalField());
	fprintf(log2file,"#Applied field (T): (%8.5f,%8.5f,%8.5f)\n",
	       bapp[0],bapp[1],bapp[2]);
      }
      // Print output columns standard (abbreviated)  header
      fprintf(log2file,"#Pass Time(ns)  Step    Torque "
	     "   Exch      Anis      Demag    Zeeman     Total "
	     "  Angle Updates\n");
      fflush(log2file); fflush(stdout);
    }
    if(pass%report_interval == 0) {
      grid->GetEnergies(exch_energy,anis_energy,demag_energy,
			zeeman_energy,total_energy);
      fprintf(log2file,
	      "%4d %7.2f  %7.4f  %7.5f %9.6f %9.6f %9.6f %9.6f %9.5f"
	      "  %5.1f %6d\n",
	      pass,timeindex*1e9,grid->GetStepSize(),last_torque,
	      exch_energy,anis_energy,demag_energy,zeeman_energy,
	      total_energy,(grid->GetMaxNeighborAngle())*180./PI,
	      grid->GethUpdateCount());
      fflush(log2file); fflush(stdout);
      if(passdumpflag) {
	grid->GetAverageMagnetization(avemag);
	if(sprintf(filename,"%s.pass%05d.vio",mif.GetOutBaseName(),pass)+1
	   >(int)sizeof(filename))
	  fatalerr("main","Overflow of filename string buffer");
	if(sprintf(winname,
		   "Pass %05d, field step %03d, m=<%6.3f,%6.3f,%6.3f>",
		   pass,
		   bstep+1,avemag[0],avemag[1],avemag[2])+1
	   >(int)sizeof(winname))
	  fatalerr("main","Overflow of winname string buffer");
	grid->BinDumpm(filename,winname);
      }
    }
    if(max_pass_count>=0 && pass>=max_pass_count) stopflag=SF_PASSMAX;
    else {
      if((last_torque=grid->StepODE())<=stop_torque) stopflag=SF_TORQUE;
      pass++; timeindex+=grid->GetTimeStep();
    }
    if(!backgroundflag && TermInputReady()) { // User interactive input
      keycode=fgetc(stdin);
      fseek(stdin,0,SEEK_END); // Skip over any extra keypresses
      printf("\r   \r"); // Erase user input
      fflush(stdout);
      switch(keycode)
	{ // Note: Telnet in ascii mode (default) won't pass Alt-chars
	case 'a': case 'A':
	  printf("Accept? (n/y)->"); fflush(stdout);
	  RawWaitTermInput(); // Wait for input
	  keycode=fgetc(stdin);
	  printf("\n");
	  if(keycode!='y') break;
	case Alt_a:
	  printf("User accepted magnetization state at pass %d\n",pass);
	  stopflag=SF_USERTORQUE;
	  break;
	case 'b': case Alt_b:
	  printf("Are you SURE you want to leave interactive mode? (n/y)->");
	  fflush(stdout);  RawWaitTermInput(); // Wait for input
	  keycode=fgetc(stdin);  printf("\n");
	  if(keycode=='y' || keycode=='Y') {
	    printf("You may re-enter interactive mode by sending a"
		   " -USR2 signal to this process (%d)\n",(int)getpid());
	    SetBackgroundMode(1);
	    sleep(2);
	  }
	  break;
	case 'd': case Alt_d:
	  // Display current magnetization state
	  printf("Display request at pass %d\n",pass);
#ifndef OLDSPRINTF
	  if(sprintf(winname,"Pass %4d; E=%f",
		     pass,grid->GetEnergy())+1>(int)sizeof(winname))
#else
	    if(strlen(sprintf(winname,"Pass %4d; E=%f",
			      pass,grid->GetEnergy()))+1>sizeof(winname))
#endif
	      fatalerr("main","Overflow of winname string buffer");
	  fflush(log2file); fflush(stdout);
	  fflush(stderr); fflush(logfile);
	  DisplayMagState(winname,grid,dolog2,log2file);
	  break;
	case Alt_f: case 'f': case 'F': // Write h to disk in vio format
#ifndef OLDSPRINTF
	  if(sprintf(winname,                   // pass Alt-chars
		     "Field h at step %d/time %g (user request)"
		     "  Energy=%f",pass,timeindex,grid->GetEnergy())+1
	     >(int)sizeof(winname))
#else
	  if(strlen(sprintf(winname,           // pass Alt-chars
		     "Field h at step %d/time %g (user request)"
		     "  Energy=%f",
		     pass,timeindex,grid->GetEnergy()))+1>sizeof(winname))
#endif
	    fatalerr("main","Overflow of winname string buffer");
#ifndef OLDSPRINTF
	  if(sprintf(filename,"delta%03dx%03dh.pass%04d.vio",Nx,Nz,pass)+1
	     >(int)sizeof(filename))
#else
	  if(strlen(sprintf(filename,"delta%03dx%03dh.pass%04d.vio",
			    Nx,Nz,pass))+1>sizeof(filename))
#endif
	    fatalerr("main","Overflow of filename string buffer");
	  printf("Writing h at pass %d in file %s\n",pass,filename);
	  grid->BinDumph(filename,winname);
	  break;
	case Alt_j: case 'j': case 'J': // Jiggle spins
	  CookedTerm();
	  printf("Enter amount to jiggle spins (default is %f)-->",JiggleSize);
	  fflush(stdout);
	  fgets(smallbuf,sizeof(smallbuf)/sizeof(smallbuf[0]),stdin);
	  if(smallbuf[0]=='\n' || smallbuf[0]=='\0') dtemp=JiggleSize;
	  else                                       dtemp=atof(smallbuf);
	  if(dtemp>0.) grid->Perturb(dtemp);
	  RawTerm();
	  break;
	case Alt_m: case 'm': case 'M': // Write m to disk in vio format
#ifndef OLDSPRINTF
	  if(sprintf(winname,                   // pass Alt-chars
		     "Magnetization at step %d/time %g (user request)"
		     "  Energy=%f",pass,timeindex,grid->GetEnergy())+1
	     >(int)sizeof(winname))
#else
	  if(strlen(sprintf(winname,           // pass Alt-chars
		     "Magnetization at step %d/time %g (user request)"
		     "  Energy=%f",
		     pass,timeindex,grid->GetEnergy()))+1>sizeof(winname))
#endif
	    fatalerr("main","Overflow of winname string buffer");
#ifndef OLDSPRINTF
	  if(sprintf(filename,"delta%03dx%03dm.pass%04d.vio",Nx,Nz,pass)+1
	     >(int)sizeof(filename))
#else
	  if(strlen(sprintf(filename,"delta%03dx%03dm.pass%04d.vio",
			    Nx,Nz,pass))+1>sizeof(filename))
#endif
	    fatalerr("main","Overflow of filename string buffer");
	  printf("Writing m at pass %d in file %s\n",pass,filename);
	  grid->BinDumpm(filename,winname);
	  break;
	case Alt_p: case 'p': case 'P': // Pause 
	  printf("Paused  <Hit any key to continue>  "); fflush(stdout);
	  RawWaitTermInput(); // Wait for input
	  keycode=fgetc(stdin);
	  printf("\n");
	  break;
	case 'q': // Note: Telnet in ascii mode (default) won't pass Alt-chars
	  printf("Stop? (n/y)->"); fflush(stdout);
	  RawWaitTermInput(); // Wait for input
	  keycode=fgetc(stdin);
	  printf("\n");
	  if(keycode!='y') break;
	case Alt_q:
	  printf("User requested abort at pass %d\n",pass);
	  stopflag=SF_USERSTOP;
	  break;
	case 'r': case Alt_r:
	  autodisplayflag^=1;
	  printf("Auto-results display turned %s\n",
		 (autodisplayflag?"on":"off"));
	  break;
	case Alt_t: case 't': case 'T': // Report usage times
	  printf("Run time usage at pass %d:\n",pass);
	  grid->PrintUsageTimes(stdout);
	  printf("#\n# Total time---\n");
	  timer.Print(stdout);
	  fflush(stdout);
	  break;
	case Alt_w: case 'w': case 'W': // Write mag. and h to disk
#ifndef OLDSPRINTF
	  if(sprintf(winname,                   // pass Alt-chars
		     "Magnetization at step %d/time %g (at user request)"
		     "  Energy=%f",pass,timeindex,grid->GetEnergy())+1
	     >(int)sizeof(winname))
#else
	  if(strlen(sprintf(winname,           // pass Alt-chars
		     "Magnetization at step %d/time %g (at user request)"
		     "  Energy=%f",
		     pass,timeindex,grid->GetEnergy()))+1>sizeof(winname))
#endif
	    fatalerr("main","Overflow of winname string buffer");
#ifndef OLDSPRINTF
	  if(sprintf(filename,"delta%03dx%03dm.pass%04d.ascii",Nx,Nz,pass)+1
	     >(int)sizeof(filename))
#else
	  if(strlen(sprintf(filename,"delta%03dx%03dm.pass%04d.ascii",
			    Nx,Nz,pass))+1>sizeof(filename))
#endif
	    fatalerr("main","Overflow of filename string buffer");
	  printf("Writing state at pass %d in file %s\n",pass,filename);
	  grid->Dump(filename,winname);
	  break;
	case EOF: // ignore
	  break;
	default:
	  fprintf(stdout,"Unknown keycode ->%c<- (%02X)\n",
		  keycode,keycode);
	  fprintf(stdout,"\n\nType---\n");
	  fprintf(stdout," [Alt-]a to Accept current state and continue"
                         " to next field step.\n");
	  fprintf(stdout," [Alt-]b to run in Batch mode.\n");
	  fprintf(stdout," [Alt-]d to Display current magnetization.\n");
	  fprintf(stdout," [Alt-]f to write current (reduced)"
		  " Field state (vio).\n");
	  fprintf(stdout," [Alt-]j to Jiggle current magnetization.\n");
	  fprintf(stdout," [Alt-]m to write current (reduced)"
		  " Magnetization state (vio).\n");
	  fprintf(stdout," [Alt-]p to Pause.\n");
	  fprintf(stdout," [Alt-]r to toggle automatic Results display %s\n",
		  (autodisplayflag?"off":"on"));
	  fprintf(stdout," [Alt-]t to show usage Times.\n");
	  fprintf(stdout," [Alt-]w to Write current state (ascii).\n");
	  fprintf(stdout," [Alt-]q to Quit.\n\n");
	  fflush(stdout);
	  sleep(2);
	}
    }
    // At local energy minimum?
    if(stopflag==SF_TORQUE
      && (grid->GetEnergy()<jiggle_energy || jiggle_count<JiggleTime)) {
      // Time to jiggle
      fprintf(log2file,
	      "# Torque too small to continue;  Magnetizations being "
	      "perturbed\n# by <%f to test width of energy minima\n",
	      JiggleSize); fflush(log2file); fflush(stdout);
      if(jiggle_count>0 && grid->GetEnergy()<jiggle_energy) { // New best state
	jiggle_count=1; jiggle_energy=grid->GetEnergy();
      }
      else jiggle_count++;
      grid->Perturb(JiggleSize);
      stopflag=SF_NOSTOP;
    }
    // Report current state to log file and display?
    if(stopflag==SF_TORQUE || stopflag==SF_USERTORQUE
       || stopflag==SF_USERSTOP || stopflag==SF_PASSMAX) {
#if ( VECDIM!=3 || MIFVECDIM!=3 )
#error VECDIM and MIFVECDIM must be equal to 3
#endif
      bapp.Dup(mif.GetNomExternalField());
      grid->GetAverageMagnetization(avemag);
      fprintf(logfile," %3d  %6d %9.2f %8.5f %8.5f %8.5f"
	      " %8.5f %8.5f %8.5f %8.5f %6.2f %7d\n",
	      bstep+1,pass,timeindex*1e9,
	      bapp[0],bapp[1],bapp[2],
	      avemag[0],avemag[1],avemag[2],grid->GetEnergy(),
	      (grid->GetMaxNeighborAngle())*180./PI,grid->GethUpdateCount());
      fflush(logfile);
      if(stopflag!=SF_PASSMAX || pass%report_interval != 0) {
	grid->GetEnergies(exch_energy,anis_energy,demag_energy,
			  zeeman_energy,total_energy);
	fprintf(log2file,
		"%4d %7.2f  %7.4f  %7.5f %9.6f %9.6f %9.6f %9.6f %9.5f"
		"  %5.1f %6d\n",
		pass,timeindex*1e9,grid->GetStepSize(),last_torque,
		exch_energy,anis_energy,demag_energy,zeeman_energy,
		total_energy,(grid->GetMaxNeighborAngle())*180./PI,
		grid->GethUpdateCount());
	fflush(log2file); fflush(stdout);
      }
      if(sprintf(filename,"%s.%03d.vio",mif.GetOutBaseName(),bstep+1)+1
	 >(int)sizeof(filename))
	fatalerr("main","Overflow of filename string buffer");
      if(sprintf(winname,"Field step %03d, m=<%6.3f,%6.3f,%6.3f>",
		 bstep+1,avemag[0],avemag[1],avemag[2])+1
	 >(int)sizeof(winname))
	fatalerr("main","Overflow of winname string buffer");
      grid->BinDumpm(filename,winname);
      if(!backgroundflag && autodisplayflag) { // Pop-up display window
	  fflush(log2file); fflush(stdout);
	  fflush(stderr); fflush(logfile);
	  DisplayMagState(winname,grid,dolog2,log2file);
      }
    }
    // Step applied field?
    if((stopflag==SF_TORQUE || stopflag==SF_USERTORQUE)
       && bstep<bsteptotal) {
      // Increment Bapplied
      mif.StepField(); bstep++;
      bapp.Dup(mif.GetNomExternalField());
      extfield->SetNomField(mif.GetMs(),bapp);
      grid->ResetODE(); // Reset ODE solver
      jiggle_count=0;
      stopflag=SF_NOSTOP;
      fprintf(log2file,
	      "# Applied field stepped to (T): (%8.5f,%8.5f,%8.5f),\n"
	      "# Pass=%5d, Max Neighbor Angle=%6.2f degrees,"
	      " Field Updates=%5d\n",bapp[0],bapp[1],bapp[2],pass,
	      (grid->GetMaxNeighborAngle())*180./PI,grid->GethUpdateCount());
      fflush(log2file); fflush(stdout);
    }
  } while(stopflag==SF_NOSTOP);
  if(stopflag==SF_PASSMAX) {
    fprintf(logfile,"# Termination on pass %d caused by"
	    " max iteration constraint\n",pass);
    fprintf(log2file,"# Termination on pass %d caused by"
	    " max iteration constraint\n",pass);
  }
  if(stopflag==SF_USERSTOP) {
    fprintf(logfile,"# Termination on pass %d caused by"
	    " user request\n",pass);
    fprintf(log2file,"# Termination on pass %d caused by"
	    " user request\n",pass);
  }
  timer.Print(logfile);
  if(!backgroundflag) ResetTerm(1);
  fprintf(log2file,"#\n# Final average magnetization:");
  avemag.Print(log2file); fprintf(log2file,"\n#\n");
  grid->PrintUsageTimes(log2file);
  fprintf(log2file,"#\n# Total time---\n");
  timer.Print(log2file);
  fflush(log2file); fflush(stdout);

  if(grid!=(Grid2D *)NULL)     delete grid;

  // NOTE: extfield is currently owned by mif, and
  //       will be deleted by it.

#ifdef TIMETEST
  fprintf(log2file,"#\n# FFT+Wrapper Time---\n");
  tempwatch[1].Print(log2file);
  fprintf(log2file,"#\n# FFT Time---\n");
  tempwatch[0].Print(log2file);
  fprintf(log2file,"#\n# Convolution Mult.---\n");
  tempwatch[2].Print(log2file);
  fprintf(log2file,"#\n# Load+Write Time---\n");
  tempwatch[3].Print(log2file);
#endif // TIMETEST

  time(&mytime);
  fprintf(logfile,"# Stop time: %s\n",ctime(&mytime));
  fclose(logfile);
  fprintf(log2file,"# Stop time: %s\n",ctime(&mytime));
  fclose(log2file);

  fprintf(stderr,"\7"); // Sound bell
  return 0;
}
