# FILE: account.tcl
#
# An object which plays the role of an account within an interpreter.
#
# Each host on the internet may have many accounts -- separate areas
# of storage each for the use of one or more users, usually with access
# controlled by some sort of security measure such as a password.  Each
# thread which runs on a host is owned by exactly one of the accounts on
# that host.  Thus, in a hierarchy of "ownership," accounts lie between
# hosts and threads.
#
# The Tcl event loop must be running in order for non-blocking socket
# communications to work.
#

em_class net_account {

    # Index of account objects by name -- used to avoid duplicate objects.
    common index {[em_array new]}

    # Name of host where the account this object represents resides
    const public variable hostname = localhost

    # Name of account this object represents
    const public variable accountname

    # Default initializer of the accountname.  Tries to determine
    # account name under which current thread is running.
    proc DefaultAccountName {} {
        global env tcl_platform
        switch -- $tcl_platform(platform) {
            unix {
                if {[info exists env(USER)]} {
                    return $env(USER)
                }
                if {[info exists env(LOGNAME)]} {
                    return $env(LOGNAME)
                }
                return oommf
            }
            windows {
                # Find some reliable entry in the registry
                return oommf
            }
            default {
                return oommf
            }
        }
    }

    # Milliseconds to wait for account thread to start on localhost
    public variable startWait = 60000
    # The ID of the current timeout (returned from 'after' command)
    private variable timeoutevent

    # The host which on which this account resides
    private variable host

    # Our connection to the actual account thread
    private variable connection

    # Port on host on which account thread listens
    private variable port
    
    # Log for logging messages
    public variable log {
        if {[catch {$log Append "Logging messages generated by $this."}]} {
            error "$log is not able to log messages."
        }
        if {[info exists logHandler]} {
            $logHandler delete
        }
        set logHandler [em_eventHandler new $log delete "$this delete"]
        if {$imadelog} {
            $oldlog delete
        }
        set oldlog $log
    }
    private variable oldlog = em_nop
    private variable logHandler
    private variable imadelog = 0

    private variable msgHandler

    # Socket for receiving ping from localhost OOMMF account server
    private variable listensocket

    # Send id
    private variable id = 0

    # Reply message
    private variable reply = {}

    # Is object ready to relay messages to and from the host?
    private variable ready = 0

    # Event handling support
    private variable bindtags
    private variable myEventHandlers = {}
    private variable rhandlers
    private variable qhandler
    private variable thandler

    private variable protocolversion

    public variable gui = 0 {
        if {$gui} {   
            if {[info exists myWidget]} {
                return
            }
            package require Tk
            set myWidget [net_accountGui new $parentWin -account $this]
            pack [$myWidget cget -winpath] -side left -fill both -expand 1
        } else {
            if {![info exists myWidget]} {
                return
            }
            $myWidget delete
            unset myWidget
        }
        upvar #0 ${this}_BtnVar btvar
        set btnvar $gui
    }    
    public variable parentWin = .
    private variable myWidget

    method ToggleGui { pwin } {
        set parentWin $pwin
        upvar #0 ${this}_BtnVar btnvar
        $this configure -gui $btnvar
    }

    # Constructor is a non-blocking operation.  After it returns an instance
    # id, the Ready method of that instance will return a boolean indicating
    # whether the object is ready.  If not, then when it becomes ready, it
    # will generate an event with id 'Ready'.  If the object cannot become
    # ready, it will delete itself, generating an event with is 'delete'.
    # Users of instances of this class should set up to handle those two
    # events, when the Ready method returns false.
    #
    constructor { args } {
        eval $this configure $args
        if {![info exists accountname]} {
            set accountname [net_account DefaultAccountName]
        }
        # Enforce only one object per account
        if {![catch {$index lookup $hostname,$accountname} existingaccount]} {
            unset hostname
            unset accountname
            $this delete
            array set argarray $args
            catch {unset argarray(-hostname)}
            catch {unset argarray(-accountname)}
            eval $existingaccount configure [array get argarray]
            return $existingaccount
        }
        $index set [list $hostname,$accountname $this]
        if {![info exists log]} {
            # Provide a default log
            $this configure -log [em_log new \
                    -title "Account $hostname:$accountname Log"]
            set imadelog 1
        }
        # Enable event generation
        set bindtags [list $this $class]
        if {[catch {net_host new -hostname $hostname -log $log} msg]} {
            error "Can't construct host object"
        } else {
            set host $msg
            lappend myEventHandlers [em_eventHandler new $host delete \
                    [list $this delete]]
            if {[$host Ready]} {
                $this LookupAccountPort
            } else {
                lappend myEventHandlers [em_eventHandler new $host Ready \
                        [list $this LookupAccountPort]]
            }
        }
    }

    method LookupAccountPort {args} {
        if {[llength $args]} {
            $log Append "$this: Received ping!"
            close [lindex $args 0]
        }
        set qid [$host Send lookup $accountname]
        set qhandler [em_eventHandler new $host Reply$qid \
                [list $this GetAccountPort]]
        set thandler [em_eventHandler new $host Timeout$qid \
                [list $this NoAnswerFromHost]]
    }

    method GetAccountPort {} {
        $qhandler delete
        unset qhandler
        $thandler delete
        unset thandler
        set reply [$host Get]
        if {[llength $reply] < 2} {
            $lig Append "Bad reply from host $hostname: $reply"
            error "Bad reply from host $hostname: $reply"
        }
        switch -- [lindex $reply 0] {
            0 {
                set port [lindex $reply 1]
                if {[info exists listensocket]} {
                    close $listensocket
                    unset listensocket
                }
                if {[info exists timeoutevent]} {
                    after cancel $timeoutevent
                    unset timeoutevent
                }
                $this Connect
            }
            default {
                # Host returns an error - assume that means lookup failed
                if {[string match localhost [$host cget -hostname]]} {
                    if {![info exists listensocket]} {
                        $this StartLocalAccount
                    }
                } else {
                    $log Append \
"No account $accountname registered with host [$host cget -hostname]"
                    $this delete
                }
            }
        }
    }

    method NoAnswerFromHost {} {
        $qhandler delete
        unset qhandler
        $thandler delete
        unset thandler
        # Host not responding -- give up.
        $log Append "No reply from host [$host cget -hostname]"
        $this delete
    }

    # Launch an OOMMF host server on localhost
    private method StartLocalAccount {} {
        set dir [[$class extension] directory]
        set acctthread [file join $dir threads account.tcl]
        if {[file readable $acctthread]} {
            # Set up to receive ping from account thread or timeout
            set listensocket [socket -server [list $this LookupAccountPort] 0]
            set listenport [lindex [fconfigure $listensocket -sockname] 2]
            set timeoutevent [after $startWait $this StartTimeout]
            $log Append \
                "Starting OOMMF account server for $accountname on localhost"
            # Ought to redirect std channels somewhere
            global omf_shell
            if {[info exists omf_shell]} {
                exec $omf_shell -notk $acctthread 0 $listenport \
                        < [em_extension NullFile] >>& [em_extension NullFile] &
            } else {
                exec [info nameofexecutable] $acctthread 0 $listenport \
                        < [em_extension NullFile] >>& [em_extension NullFile] &
            }
        } else {
            error "Install error: No OOMMF account server available to start"
        }
    }

    method StartTimeout {} {
        $log Append "Timed out waiting for OOMMF account server to start for 
	$accountname on localhost"
        $this delete
        return
    }

    method Connect {} {
        if {[catch {
                net_connection new -hostname $hostname -port $port -log $log
                } msg]} {
            $this delete
        } else {
            set connection $msg
            lappend myEventHandlers [em_eventHandler new $connection delete \
                    [list $this delete]]
            lappend myEventHandlers [em_eventHandler new $connection Ready \
                    [list $this ConnectionReady]]
        }
    }

    method ConnectionReady {} {
        set msgHandler [em_eventHandler new $connection Readable \
                [list $this VerifyOOMMFAccount]]
    }

    method VerifyOOMMFAccount {} {
        $msgHandler delete
        unset msgHandler
        if {![regexp {OOMMF account protocol ([0-9.]+)} [$connection Get] \
                match version]} {
            # Not an OOMMF account
            $log Append "No OOMMF account server answering on $hostname:$port"
            $this delete
            return
        }
        set protocolversion $version
        set rhandlers [em_array new]
        $log Append "OOMMF account $hostname:$accountname ready"
        set msgHandler [em_eventHandler new $connection Readable \
                [list $this Hear]]
        set ready 1
        em_eventHandler Generate Ready $bindtags
    }

    # Handle 'tell' messages from connection
    method Hear {} {
        set reply [$connection Get]
        em_eventHandler Generate Readable $bindtags
    }

    method Receive { rid } {
        $rhandlers unset $rid
        set reply [$connection Get] 
        em_eventHandler Generate Reply$rid $bindtags
    }

    method Get {} {
        return $reply
    }

    # Send a command to the host
    method Send { args } {
        if {!$ready} {
            error "host $hostname not ready"
        }
        incr id
        set qid [eval $connection Query $args]
        $rhandlers set [list $id [em_eventHandler new $connection Reply$qid \
                [list $this Receive $id]]]
        return $id
    }

    method Ready {} {
        return $ready
    }

    destructor {
        set ready 0
        if {[info exists bindtags]} {
            em_eventHandler Generate delete $bindtags
        }
        # delete handlers before deleting event-generating members
        foreach handler $myEventHandlers {
            $handler delete
        }
        if {[info exists rhandlers]} {
            foreach idx [$rhandlers names] {
                [$rhandlers lookup $idx] delete
            }
            $rhandlers delete
        }
        if {[info exists thandler]} {
            $thandler delete
        }
        if {[info exists qhandler]} {
            $qhandler delete
        }
        if {[info exists msgHandler]} {
            $msgHandler delete
        }
        if {[info exists connection]} {
            $connection delete
        }
        if {[info exists timeoutevent]} {
            after cancel $timeoutevent
        }
        if {[info exists listensocket]} {
            close $listensocket
        }
        if {[info exists log]} {
            $log Append "Deleting $this ..."
            $this configure -log em_nop
        }
        if {[info exists myWidget]} {
            $this configure -gui 0
        }
        if {[info exists hostname] && [info exists accountname]} {
            $index unset $hostname,$accountname
        }
    }

}

