# FILE: thread.tcl
#
# An object which plays the role of a thread within an interpreter.
#
# The Tcl event loop must be running in order for non-blocking socket
# communications to work.
#

em_class net_thread {

    # Index of thread objects by hostname,pid
    common index {[em_array new]}

    # Name of host where the account this object represents resides
    const public variable hostname = localhost

    # Name of account this object represents
    const public variable accountname

    # Process ID of the thread
    const public variable pid

    # The "alias" of this thread
    const public variable alias

    # The account under which this thread is running
    private variable account

    # Our connection to the actual running thread
    private variable connection

    # Port on host on which the thread server listens
    private variable port
    
    # Log for logging messages
    public variable log {
        if {[catch {$log Append "Logging messages generated by $this."}]} {
            error "$log is not able to log messages."
        }
        if {[info exists logHandler]} {
            $logHandler delete
        }
        set logHandler [em_eventHandler new $log delete "$this delete"]
        if {$imadelog} {
            $oldlog delete
        }
        set oldlog $log
    }
    private variable oldlog = em_nop
    private variable logHandler
    private variable imadelog = 0

    private variable msgHandler

    # Send id
    private variable id = 0

    # Reply message
    private variable reply

    # Is object ready to relay messages to and from the thread?
    private variable ready = 0

    # Event handling support
    private variable bindtags
    private variable myEventHandlers = {}
    private variable rhandlers
    private variable thandlers
    private variable qhandler
    private variable thandler

    private variable protocolname
    private variable protocolversion

    method Protocol {} {
        return $protocolname
    }

    public variable gui = 0 {
        if {$gui} {   
            if {[info exists myWidget]} {
                return
            }
            package require Tk
            set myWidget [net_threadGui new -thread $this]
        } else {
            if {![info exists myWidget]} {
                return
            }
            $myWidget delete
            unset myWidget
        }
        upvar #0 ${this}_BtnVar btvar
        set btnvar $gui
    }    
    private variable myWidget
    private variable hasGui = 0

    method HasGui {} {
        return $hasGui
    }
 
    method ToggleGui { } {
        upvar #0 ${this}_BtnVar btnvar
        $this configure -gui $btnvar
    }

    # Constructor is a non-blocking operation.  
    constructor { args } {
        eval $this configure $args
        if {![info exists accountname]} {
            set accountname [net_account DefaultAccountName]
        }
        if {![info exists pid]} {
            error "Thread must have pid!"
        }
        # Enforce only one object per account
        if {![catch {$index lookup $hostname,$pid} existingthread]} {
            unset hostname
            unset pid
            $this delete
            array set argarray $args
            catch {unset argarray(-hostname)}
            catch {unset argarray(-accountname)}
            catch {unset argarray(-pid)}
            catch {unset argarray(-alias)}
            eval $existingthread configure [array get argarray]
            return $existingthread
        }
        $index set [list $hostname,$pid $this]
        if {![info exists log]} {
            # Provide a default log
            $this configure -log [em_log new \
                    -title "Thread $hostname:$pid Log"]
            set imadelog 1
        }
        # Enable event generation
        set bindtags [list $this $class]
        if {[catch {
                net_account new -hostname $hostname -log $log \
                -accountname $accountname} msg]} {
            error "Can't construct account object"
        } else {
            set account $msg
            lappend myEventHandlers [em_eventHandler new $account delete \
                    [list $this delete]]
            if {[$account Ready]} {
                $this LookupThreadPort
            } else {
                lappend myEventHandlers [em_eventHandler new $account Ready \
                        [list $this LookupThreadPort]]
            }
        }
    }

    method LookupThreadPort {} {
        set qid [$account Send lookup $pid]
        set qhandler [em_eventHandler new $account Reply$qid \
                [list $this GetThreadPort]]
        set thandler [em_eventHandler new $account Timeout$qid \
                [list $this NoAnswerFromAccount]]
    }

    method GetThreadPort {} {
        $qhandler delete
        unset qhandler
        $thandler delete
        unset thandler
        set reply [$account Get]
        if {[llength $reply] < 2} {
            $lig Append "Bad reply from account $accountname: $reply"
            error "Bad reply from account $accountname: $reply"
        }
        switch -- [lindex $reply 0] {
            0 {
                set port [lindex $reply 1]
                $this Connect
            }
            default {
                # Account returns an error - assume that means lookup failed
                $log Append \
"No thread $pid registered with account [$account cget -accountname]"
                $this delete
            }
        }
    }

    method NoAnswerFromAccount {} {
        $qhandler delete
        unset qhandler
        $thandler delete
        unset thandler
        # Host not responding -- give up.
        $log Append "No reply from account [$account cget -accountname]"
        $this delete
    }

    method Connect {} {
        if {[catch {
                net_connection new -hostname $hostname -port $port -log $log
                } msg]} {
            $this delete
        } else {
            set connection $msg
            lappend myEventHandlers [em_eventHandler new $connection delete \
                    [list $this delete]]
            lappend myEventHandlers [em_eventHandler new $connection Ready \
                    [list $this ConnectionReady]]
        }
    }

    method ConnectionReady {} {
        set msgHandler [em_eventHandler new $connection Readable \
                [list $this VerifyOOMMFThread]]
    }

    method VerifyOOMMFThread {} {
        $msgHandler delete
        unset msgHandler
        if {![regexp {OOMMF ([^ ]+) protocol ([0-9.]+)} [$connection Get] \
                match name version]} {
            # Not an OOMMF thread
            $log Append "No OOMMF thread server answering on $hostname:$port"
            $this delete
            return
        }
        set protocolname $name
        set protocolversion $version
        set rhandlers [em_array new]
        set thandlers [em_array new]
        $log Append "OOMMF thread $hostname:$pid ready"
        set msgHandler [em_eventHandler new $connection Readable \
                [list $this Hear]]
        set ready 1
        incr id
        set qid [$connection Query HasGui]
        $rhandlers set [list $id [em_eventHandler new $connection Reply$qid \
                [list $this ReceiveHasGui $id]]]
        $thandlers set [list $id [em_eventHandler new $connection Timeout$qid \
                [list $this Timeout $id]]]
    }

    method ReceiveHasGui { rid } {
        [$rhandlers lookup $rid] delete
        [$thandlers lookup $rid] delete
        $rhandlers unset $rid
        $thandlers unset $rid
        set reply [$connection Get] 
        if {[lindex $reply 0]} {
            error "error in reply to HasGui query"
        } else {
            set hasGui [lindex $reply 1]
        }
        em_eventHandler Generate Ready $bindtags
    }

    # Handle 'tell' messages from connection
    method Hear {} {
        set reply [$connection Get]
        em_eventHandler Generate Readable $bindtags
    }

    method Receive { rid } {
        [$rhandlers lookup $rid] delete
        [$thandlers lookup $rid] delete
        $rhandlers unset $rid
        $thandlers unset $rid
        set reply [$connection Get] 
        em_eventHandler Generate Reply$rid $bindtags
    }

    method Get {} {
        return $reply
    }

    # Send a command to the thread server
    method Send { args } {
        if {!$ready} {
            error "thread $hostname:$pid not ready"
        }
        incr id
        set qid [eval $connection Query $args]
        $rhandlers set [list $id [em_eventHandler new $connection Reply$qid \
                [list $this Receive $id]]]
        $thandlers set [list $id [em_eventHandler new $connection Timeout$qid \
                [list $this Timeout $id]]]
        return $id
    }

    method Timeout { rid } {
        [$rhandlers lookup $rid] delete
        [$thandlers lookup $rid] delete
        $rhandlers unset $rid
        $thandlers unset $rid
    }

    method Ready {} {
        return $ready
    }

    destructor {
        set ready 0
        if {[info exists bindtags]} {
            em_eventHandler Generate delete $bindtags
        }
        # delete handlers before deleting event-generating members
        foreach handler $myEventHandlers {
            $handler delete
        }
        if {[info exists rhandlers]} {
            foreach idx [$rhandlers names] {
                [$rhandlers lookup $idx] delete
            }
            $rhandlers delete
        }
        if {[info exists thandlers]} {
            foreach idx [$thandlers names] {
                [$thandlers lookup $idx] delete
            }
            $thandlers delete
        }
        if {[info exists thandler]} {
            $thandler delete
        }
        if {[info exists qhandler]} {
            $qhandler delete
        }
        if {[info exists msgHandler]} {
            $msgHandler delete
        }
        if {[info exists connection]} {
            $connection delete
        }
        if {[info exists log]} {
            $log Append "Deleting $this ..."
            $this configure -log em_nop
        }
        if {[info exists myWidget]} {
            $this configure -gui 0
        }
        if {[info exists hostname] && [info exists pid]} {
            $index unset $hostname,$pid
        }
    }

}

