# FILE: entrybox.tcl
#
# Entry box dialog sub-widget
#
# debugging enable/disable/level

#   Use the method call "ReadEntryString" to get the value
# currently displayed in the entry box.  Use "cget -value" to
# get the last value "committed" with a <Key-Return> event.
#   The <Key-Return> event copies the value from the entry widget
# into -value, and then calls the -callback proc, with ACTIONID of
# UPDATE, and arg list of -value $value.
#   Set -autoundo if you want automatic undo action (equivalent
# to the <Key-Escape> event) on <FocusOut> events.  This copies the
# last "committed" value (as stored in -value) into the entry widget
# display.
#   Use the method SetEntryString to set both -value and the displayed
# text.
em_class omf_entrybox {
    const public variable winpath
    const public variable outer_frame_options = "-bd 2 -relief raised"
    private variable winentry
    const public variable label = {}
    const public variable labelwidth = 0
    public variable callback = em_nop
    const public variable valuewidth = 16
    public variable value = {}
    public variable display_preference = {} ;# Index of entry string to
    ## make visible on external updates.  Typically set to "end" for
    ## potentially long strings like pathnames.
    const public variable autoundo = 0 ;# If set, then widget performs
    ## an undo action on FocusOut events
    constructor {basewinpath args} {
	# Don't add '.' suffix if there already is one.
	# (This happens chiefly when basewinpath is exactly '.')
	if {![string match {.} \
		[string index $basewinpath \
		[expr [string length $basewinpath]-1]]]} {
	    append basewinpath {.}
	}
	eval $this configure -winpath ${basewinpath}w$this $args

	# Temporarily disable callbacks until construction is complete
	set callback_keep $callback ;	set callback em_nop

	# Pack entire subwidget into a subframe.  This makes
	# housekeeping easier, since the only bindings on
	# $winpath will be ones added from inside the class,
	# and also all subwindows can be destroyed by simply
	# destroying $winpath.
	eval frame $winpath $outer_frame_options
	label $winpath.label -text $label
	if {[string compare $labelwidth ""]!=0} {
	    $winpath.label configure -width $labelwidth -anchor e
	}
	pack $winpath.label -side left -anchor e

	set winentry [entry $winpath.value -relief sunken -bd 2 \
		-width $valuewidth -bg [$winpath.label cget -bg]]
	## The default color scheme under Tcl 8.0 + Windows makes
	## labels a light gray, and entry backgrounds bright white.
	## *I* think it looks pretty hideous...  -mjd 97/10/3
	pack $winentry -fill x -expand 1 -side left
	$this SetEntryString $value

	# Widget bindings
	bind $winentry <Key-Return> "$this CommitValue"
	bind $winentry <Key-Escape> "$this Undo"
	if {$autoundo} {
	    bind $winentry <FocusOut> [bind $winentry <Key-Escape>]
	}
	bind $winpath <Destroy> "$this WinDestroy %W"

	# Re-enable callbacks
	set callback $callback_keep
    }
    method ReadEntryString {} { winentry } {
	return [$winentry get]
    }
    method CommitValue {} { winentry callback value } {
	set value [$winentry get]
	eval $callback $this UPDATE -value {$value}
    }
    method SetEntryString { newvalue } { winentry value display_preference } {
	if {[string compare $newvalue [$winentry get]]!=0} {
	    $winentry delete 0 end
	    $winentry insert 0 $newvalue
	}
	if {![string match {} $display_preference]} {
	    $winentry xview $display_preference
	}
	set value $newvalue
    }
    method Undo {} { value } {
	$this SetEntryString $value
    }
    method WinDestroy { w } { winpath } {
	if {[string compare $winpath $w]!=0} {
	    return    ;# Child destroy event
	}
	$this delete
    }
    destructor {
	if {[winfo exists $winpath]} {
            # Remove <Destroy> binding
	    bind $winpath <Destroy> {}
	    # Destroy instance frame, all children, and bindings
	    destroy $winpath
	}
    }
}

