# $Name:  $ 
# $Author: dgp $ 
# $Date: 1997/12/18 20:41:59 $ 
#
# The startup file for the Tcl application Oommf, the Object-Oriented
# MicroMagnetic computational Framework.  Its purpose is to establish
# a suitable Tcl environment for the Oommf application.
#
# In particular, it performs the following tasks
#	o	Makes Oommf's support packages known to the auto-loader
#	o	Checks that the versions of the support packages and the
#		binary are compatible
#	o	Defines the Tcl command 'package' if it is missing (pre-7.5)
#	o	Configures the behavior of the package unknown handler to be
#		consistent with Tcl 7.6 .
#	o	Defines the standard procs and global Tcl variables for
#		the Oommf package.

# Re-define some basic procs, which were already defined in 
# oommfInit.cc .  This lets us source this file from wish.
proc OMF_PackageName {} {return Oommf}
proc OMF_FileNameBase {} {string tolower [OMF_PackageName]}

#bgerror "Starting oommf.tcl!"
# Add the directory named by the OOMMF_PACKAGE_PATH environment variable
# to the path searched by the auto-loader.
#
# This allows users to replace Oommf's support packages with their own 
# replacement implementations.
catch {lappend auto_path $env([string toupper [OMF_PackageName]]_PACKAGE_PATH)}

# Add the directory containing default implementations Oommf's support
# packages to the path searched by the auto_loader.
if {[lsearch -exact $auto_path $omf_library] < 0} {
    lappend auto_path $omf_library
}

# This version of Oommf is written for Tcl/Tk versions 7.5/4.1 or later
if {[catch {package require Tcl 8}]} {
  package require Tcl 7.5
}
#if {[catch {package require Tk 8}]} {
#  package require Tk 4.1
#}

# The default package unknown handlers included in versions of Tcl earlier
# than 8.0 have a great weakness. If there is an error in *any* pkgIndex.tcl
# file in the current search path of the package unknown command, that
# error will bring the application to a halt -- even if the error is in
# an index file provided by an entirely different application!  The default
# package unknown handler included in Tcl 8.0 is far more robust, and is
# compatible back to Tcl 7.5.  It is reproduced below and made the 
# package unknown handler for OOMMF applications. 
#
# Another benefit to this is the avoidance of the inconsistency between
# Tcl 7.5 and Tcl 7.6+ about what constitutes the search path for the
# package unknown handler.
#
# NOTE: Because this *replaces* the default package unknown handler,
# it is possible that a specially customized installation of Tcl in
# which a specialized package unkown handler has already been defined
# and is required to handled the customized environment may make that
# customization mechanism broken by this strategy.  Too bad.  OOMMF
# goes best with plain Tcl/Tk.
#
if {$tcl_version < 8} {
  set _omf_g(defaultPkgUnknown) [package unknown]

  proc _OMF_PkgUnknown {name version {exact {}}} {
    global auto_path tcl_platform env dir

    if ![info exists auto_path] {
	return
    }
    if {[info exists dir]} {
	set save_dir $dir
    }
    for {set i [expr [llength $auto_path] - 1]} {$i >= 0} {incr i -1} {
	# we can't use glob in safe interps, so enclose the following
	# in a catch statement
	catch {
	    foreach file [glob -nocomplain [file join [lindex $auto_path $i] \
		    * pkgIndex.tcl]] {
		set dir [file dirname $file]
		if [catch {source $file} msg] {
		    tclLog "error reading package index file $file: $msg"
		}
	    }
        }
	set dir [lindex $auto_path $i]
	set file [file join $dir pkgIndex.tcl]
	# safe interps usually don't have "file readable", nor stderr channel
	if {[interp issafe] || [file readable $file]} {
	    if {[catch {source $file} msg] && ![interp issafe]}  {
		tclLog "error reading package index file $file: $msg"
	    }
	}
	# On the Macintosh we also look in the resource fork 
	# of shared libraries
	# We can't use tclMacPkgSearch in safe interps because it uses glob
	if {(![interp issafe]) && ($tcl_platform(platform) == "macintosh")} {
	    set dir [lindex $auto_path $i]
	    tclMacPkgSearch $dir
		foreach x [glob -nocomplain [file join $dir *]] {
		    if [file isdirectory $x] {
			set dir $x
			tclMacPkgSearch $dir
		    }
		}
	}
    }
    if {[info exists save_dir]} {
	set dir $save_dir
    } else {
	unset dir
    }
  }
  package unknown _OMF_PkgUnknown
}

# Patch a bug MJD found in the Tk 4.2 distribution
if { [info exists tk_version] && ($tk_version == 4.2) } {
    # The distribution tkMenuFind source seems to be broken: it has
    # comments in an illegal position inside the 'switch' statement.
    # The following code is from Tk 4.2, but with the comments moved.
    proc tkMenuFind { w char } {
	global tkPriv
	set char [string tolower $char]
	
	foreach child [winfo child $w] {
	    switch [winfo class $child] {
		Menubutton {
		    set char2 [string index [$child cget -text]  \
			    [$child cget -underline]]
		    if {([string compare $char \
			    [string tolower $char2]] == 0) \
			    || ($char == "")} {
			if {[$child cget -state] != "disabled"} {
			    return $child
			}
		    }
		}
		default {
		    # The tag above used to be "Frame", but it was changed so
		    # that the code would work with Itcl 2.0, which apparently
		    # uses other classes of widgets to hold menubuttons.
		    set match [tkMenuFind $child $char]
		    if {$match != ""} {
			return $match
		    }
		}
	    }
	}
	return {}
    }
}

# Now that the package command conforms to Tcl 8.0+ behavior, use it.
package require em
package require oommf

#bgerror "Patches in!"
# We want to use the CvsTools package (defines OMFCVS_PackageProvide)
#bgerror "Require CvsTools!"
package require CvsTools 0

#bgerror "Provide Oommf!"
# Adds Oommf package to database; defines standard global variables
OMFCVS_PackageProvide {^Name: Oommf-0_0b0^} {}

#bgerror "Require HelpBrowser!"
# At this point list additional package requirements of the Oommf application
package require HelpBrowser 0

# Globals should be set now, so define Gui startup
proc _OMF_StartGui {} {
  global omf_library 
  uplevel #0 source [file join $omf_library [OMF_FileNameBase][OMF_Version] gui.tcl]
}

